/**********************************************************************
 *
 *  What you need to write your own primitive widget?
 *   
 *  1.  Need a TYPE (>65536) and 5 procedures.
 *       setup       (EZ_Widget *widget);
 *       computeSize (EZ_Widget *widget,  int *w, int *h);
 *       freeData    (EZ_Widget *widget);
 *       draw        (EZ_Widget *widget);
 *       eventHandle (EZ_Widget *widget, XEvent *event)
 *  
 *  2.  Call EZ_RegisterPrimitiveWidget(type, iname, cname, the_5_procedures)
 *
 *  After a widget has been registered, 
 *  use EZ_CreateWidget with type=TYPE to create a 
 *  widget.
 *
 *********************************************************************/
/************************* Locator Widget ****************************/
#include "EZ.h"

void ez_GetLocatorXY (EZ_Widget *widget, float *x, float *y);
void ez_SetLocatorXY (EZ_Widget *widget, float x,  float y);
/*********************************************************************/

#define Locator_TYPE    65537 
static void LocatorSetup       (EZ_Widget *widget);
static void LocatorFreeData    (EZ_Widget *widget);
static void LocatorComputeSize (EZ_Widget *widget, int *w, int *h);
static void LocatorDraw        (EZ_Widget *widget);
static void LocatorEventHandle (EZ_Widget *widget, XEvent *event);
/*********************************************************************/

static void LocatorSetup(EZ_Widget *widget)
{
  /* this procedure is invoked when a new LOCATOR widget
   * is created. We initialize the locator through 
   * resources and create a gc for rendering the crosshair.
   *      
   * primitive widgets are implemented using the RAW_XWINDOW
   * widget. The RAW_XWINDOW widget data structure has
   * an array of 10 EZ_UnknownDataType reserved for 
   * us to hook up private data. 
   */
  EZ_UnknownDataType  *data = EZ_GetRawXWinPrivateData(widget);
  unsigned long       mask  = 0L;  
  XGCValues           gcvalues;
  char                *value;   /* tmp return for resource */

  /* data is an array of 10 EZ_UnknownDataTypes */
  data[0].f = 0.5;
  data[1].f = 0.5;
  /* parse private resources */
  if(EZ_GetWidgetResource(widget, "value", &value))
    {
      float   x,y;      
      if(sscanf(value, "%f %f", &x, &y) == 2)
        {
          if( x >= 0.0 && x <= 1.0) data[0].f = x;
          if( y >= 0.0 && y <= 1.0) data[1].f = y;
        }
    }
  data[2].p = (void *)EZ_GetGC(mask, &gcvalues);
}
/***************************************************************/

static void LocatorFreeData(EZ_Widget *widget) 
{
  /* this procedure is called when a LOCATOR widget is
   * destroyed. We free the GC allocated for this widget 
   */
  EZ_UnknownDataType  *data = EZ_GetRawXWinPrivateData(widget);
  GC gc = (GC)data[2].p;
  /* for gc created by EZ_GetGC, you must use EZ_FreeGC to free it ! */
  EZ_FreeGC(gc); 
}
/***************************************************************/

static void LocatorComputeSize(EZ_Widget *widget, int *w, int *h)
{
  /* this procedure is invoked by the geometry manager to
   * figure out the minimal size of LOCATOR.
   */
  *w =100; *h =100;  
}
/***************************************************************/

static void LocatorDraw(EZ_Widget *widget)
{
  int focus_pad = 0;   /* kbd focus highlight padding */
  int border    = 0;   /* border style */
  int bwidth    = 0;   /* border width */
  int padx, pady;      /* XY padding, not used */
  int width, height;   /* widget window width and height */
  Display  *dpy;       /* the display */
  Window   win;        /* widget window */
  int      depth;      /* depth of visual */
  int      tile;       /* is bg tiled ? */
  int      x, y, tmp, tmp2;
  Pixmap   pixmap;
  GC       gc;
  unsigned long pv;
  EZ_UnknownDataType  *data = EZ_GetRawXWinPrivateData(widget);

  EZ_GetWidgetPaddings(widget, &padx, &pady, &focus_pad); /* get all paddings       */
  EZ_GetWidgetBorder(widget, &bwidth, &border);           /* border characteristics */
  EZ_GetWidgetDimension(widget, &width, &height);         /* widget size            */
  dpy = EZ_GetDisplay();              /* display   */
  depth = EZ_GetDepth();             /*   visual    */
  win = EZ_GetWidgetWindow(widget); /* widget window */
  tmp = focus_pad + bwidth;        
  tmp2 = tmp + tmp;

  /* draw into a tmp pixmap, copy to the widget window when done */
  pixmap = XCreatePixmap(dpy, win, width, height, depth); 

  /* fill the background */
  EZ_GetBackgroundTileGC(widget, &gc, &pv, 0,0);  
  XFillRectangle(dpy, pixmap, gc, 0, 0, width, height);

  /* fix kbd focus highlight border */
  EZ_FixFocusHighlightBorder(widget, pixmap, width, height, focus_pad);
  
  /* render the foreground, the crosshair */
  gc = (GC)data[2].p;
  pv = EZ_GetForegroundPV(widget);
  XSetForeground(dpy, gc, pv);	

  x = tmp + (int)((width - tmp2) * data[0].f + 0.5);
  y = tmp + (int)((height- tmp2) * data[1].f + 0.5);
  
  XDrawLine(dpy, pixmap, gc, x, tmp, x, height -tmp); 
  XDrawLine(dpy, pixmap, gc, tmp, y, width - tmp,y);

  /* render the widget border */
  EZ_DrawRectBorder(widget, pixmap);

  /* and finally  ... */
  XCopyArea(dpy,pixmap,win, gc,0,0,width,height,0,0); 
  XFreePixmap(dpy, pixmap); 
}
/***************************************************************/

static void LocatorEventHandle(EZ_Widget *widget, XEvent *event)
{
  if(event->type == Expose) LocatorDraw(widget);
  else if(event->type == ButtonPress || event->type == KeyPress)
    {
      int padx, pady, focus_pad=0;                /* paddings  */
      int border = 0, bwidth = 0, bwpad, bwpad2;  /* border .. */
      int x,y, width, height;                     /* dimension */
      Display *dpy = EZ_GetDisplay();             /* display   */
      XEvent xevent;
      float save_x, save_y;
      EZ_UnknownDataType  *data = EZ_GetRawXWinPrivateData(widget);

      if(event->type == ButtonPress)
        {
          if(event->xbutton.button == EZ_Btn1)
            {
              /* get paddings and  dimension */
              EZ_GetWidgetPaddings(widget, &padx, &pady, &focus_pad);
              EZ_GetWidgetBorder(widget, &bwidth, &border);
              EZ_GetWidgetDimension(widget, &width, &height);
              bwpad = focus_pad + bwidth;
              bwpad2 = bwpad + bwpad;
              x = event->xbutton.x;
              y = event->xbutton.y;
              save_x = data[0].f;
              save_y = data[1].f;
              while(1)
                {
                  /* loop and wait for ButtonRelease. Invoke
                   * private event handlers on other events. Handle
                   * MotionNotify events and execute MotionCallbacks
                   * if needed.
                   */
                  if(x >= 0)
                    {
                      data[0].f = (float)(x-bwpad)/(float)(width  - bwpad2);
                      data[1].f = (float)(y-bwpad)/(float)(height - bwpad2);
                      if(data[0].f < 0.0) data[0].f = 0.0; else if(data[0].f > 1.0) data[0].f=1.0;
                      if(data[1].f < 0.0) data[1].f = 0.0; else if(data[1].f > 1.0) data[1].f=1.0;
                      LocatorDraw(widget);
                      /* if locator has changed it state, invoke motion callbacks */
                      if(save_x != data[0].f || save_y != data[1].f)
                        EZ_CallWidgetMotionCallbacks(widget);
                    }
                  XNextEvent(dpy, &xevent);
              
                  /* tricky! We want to give private EHandlers a chance  */
                  if(EZ_FilterEvent(&xevent)) EZ_InvokePrivateEventHandler(&xevent);
                  if(xevent.type == Expose) 
                    {
                      EZ_WidgetDispatchEvent(&xevent);
                      EZ_RemoveEvent(&xevent);
                    }
                  if(xevent.type == MotionNotify)
                    {
                      x = xevent.xmotion.x;
                      y = xevent.xmotion.y;
                      while(XCheckTypedEvent(dpy, MotionNotify, &xevent));
                    }
                  else if(xevent.type == ButtonRelease && xevent.xbutton.button == Button1)
                    break;
                  else x = -1111;
                }
            } 
          else if(event->xbutton.button == EZ_Btn3) 
            {
              EZ_HandleDnDEvents(widget, event);
            }
        }
      else /* key press */
        {
          char              tmpbuffer[8];
          int               buffersize = 8, xmove_unit, ymove_unit, count;
          KeySym            keysym;
          XComposeStatus    compose; 

          count = XLookupString(&(event->xkey), tmpbuffer, buffersize, &keysym, &compose);
          tmpbuffer[count] = '\0';
          xmove_unit = 0;
          ymove_unit = 0;

          switch(keysym)
            {
            case XK_Left: case XK_b:  case XK_h: case XK_B:  case XK_H:  
#ifdef XK_KP_Left
            case XK_KP_Left:
#endif
              if(event->xkey.state & ShiftMask)        xmove_unit = -4;
              else if(event->xkey.state & Mod1Mask)    xmove_unit = -10 ;
              else if(event->xkey.state & ControlMask) xmove_unit = -20;
              else xmove_unit = -1;
              break;
            case XK_Right: case XK_f: case XK_l: case XK_F: case XK_L:  
#ifdef XK_KP_Right
            case XK_KP_Right:
#endif
              if(event->xkey.state & ShiftMask)         xmove_unit = 4;
              else if(event->xkey.state & Mod1Mask)     xmove_unit = 10 ;
              else if(event->xkey.state & ControlMask)  xmove_unit = 20;
              else xmove_unit = 1;
              break;
            case XK_Up: case XK_k:  case XK_p: case XK_K:  case XK_P: 
#ifdef XK_KP_Up
            case XK_KP_Up:
#endif
              if(event->xkey.state & ShiftMask)         ymove_unit = -4;
              else if(event->xkey.state & Mod1Mask)     ymove_unit = -10 ;
              else if(event->xkey.state & ControlMask)  ymove_unit = -20;
              else ymove_unit = -1;
              break;
              break;
            case XK_Down: case XK_n: case XK_j: case XK_N: case XK_J:  
#ifdef XK_KP_Down
            case XK_KP_Down:
#endif
              if(event->xkey.state & ShiftMask)         ymove_unit = 4;
              else if(event->xkey.state & Mod1Mask)     ymove_unit = 10 ;
              else if(event->xkey.state & ControlMask)  ymove_unit = 20;
              else ymove_unit = 1;
              break;
            default:
              break;
            }
              
          if(xmove_unit | ymove_unit)
            {
              data[0].f += (xmove_unit) * 0.01;
              data[1].f += (ymove_unit) * 0.01;
              if(data[0].f < 0.0) data[0].f = 0.0; else if(data[0].f > 1.0) data[0].f=1.0;
              if(data[1].f < 0.0) data[1].f = 0.0; else if(data[1].f > 1.0) data[1].f=1.0;
              LocatorDraw(widget);
              EZ_CallWidgetMotionCallbacks(widget);
            }
          /* normally there are too many KeyPress events ... */
          while(XCheckTypedEvent(dpy, KeyPress, &xevent));
        }
    }
}
/**********************************************************************/

void ez_GetLocatorXY(EZ_Widget *widget, float *x, float *y)
{
  /* return the coordinates of the locator */
  EZ_UnknownDataType  *data = EZ_GetRawXWinPrivateData(widget);
  *x = data[0].f;  *y = data[1].f; 
}
/**********************************************************************/

void ez_SetLocatorXY(EZ_Widget *widget, float x, float y)
{
  /* set the location of the locator */
  EZ_UnknownDataType  *data = EZ_GetRawXWinPrivateData(widget);
  if(x < 0.0) x = 0.0; else if(x > 1.0) x = 1.0;
  if(y < 0.0) y = 0.0; else if(y > 1.0) y = 1.0;
  if(x != data[0].f || y != data[1].f)
    {
      data[0].f = x;
      data[1].f = y;
      if(EZ_WidgetIsViewable(widget)) EZ_DrawWidget(widget);
      EZ_CallWidgetMotionCallbacks(widget);
    }
}
/*********************************************************************
 *
 *  Testing .....
 *
 *********************************************************************/
/*#ifdef TEST_LOCATOR*/
static EZ_Widget *copy, *copy2;

static void pEHandler(EZ_Widget *widget, void *data, int etype, XEvent *xev)
{
  int update = 0;
  switch(xev->type)
    {
    case ButtonPress:
      if(xev->xbutton.button == Button1) update = 1;
      break;
    case MotionNotify:
    case KeyPress:
      update = 1;
    default:
      break;
    }
  if(update)
    {
      float       x,y;
      char        str[128];
      EZ_Widget   *label = data;
      ez_GetLocatorXY(widget, &x, &y);
      ez_SetLocatorXY(copy, x,y);
      sprintf(str, " %f %f ", x, y);
      EZ_ConfigureWidget(label, EZ_LABEL_STRING, str, 0);
    }
}

static void motionCb(EZ_Widget *widget, void *data)
{
  float x,y;
  ez_GetLocatorXY(widget, &x, &y);
  ez_SetLocatorXY(data, x, y);
}

static void getout(EZ_Widget *widget, void *dat)
{ EZ_Shutdown();  exit(0); }


static int encodeLoc(EZ_Widget *widget, void *data, char **message, int *length, int *needFree)
{
  static char str[64];
  float x, y;
  ez_GetLocatorXY(widget, &x, &y);
  sprintf(str, " %f %f ", x, y);
  *length = strlen(str);
  *message = str;
  *needFree = 0;
  return(EZ_DND_SUCCESS);  
}
static int setDragIconEncoder(EZ_Widget *widget, void *data, char **message, int *len, int *nFree)
{
  EZ_SetDragIcon((EZ_Widget *)data, 1);
  return(EZ_DND_SUCCESS);
}
int decodeLoc(EZ_Widget *widget, void *data,  char *message, int length)
{
  printf("[%s]\n", message);
  return(EZ_DND_SUCCESS);
}

main(int ac, char **av)
{
  EZ_Widget *frame, *label, *locator, *btn, *icon;
  Atom TEST_ATOM;

  EZ_Initialize(ac,av,0);
  TEST_ATOM = EZ_GetAtom("TEST_ATOM");
  EZ_RegisterPrimitiveWidget(Locator_TYPE, "locator", "Locator",
                             LocatorSetup, LocatorComputeSize,
                             LocatorDraw, LocatorFreeData, LocatorEventHandle);
  
  frame    = EZ_CreateWidget(EZ_WIDGET_FRAME,   NULL,
                             EZ_FILL_MODE,      EZ_FILL_BOTH,
                             EZ_LABEL_STRING,   "Locator", 
                             EZ_ORIENTATION,    EZ_VERTICAL,
                             NULL);
  locator  = EZ_CreateWidgetXrm(Locator_TYPE,    frame,
                                "locator1",         "Locator1",
                                EZ_BORDER_TYPE,  EZ_BORDER_DOWN,
                                EZ_PADB,         2,
                                EZ_BORDER_WIDTH, 4,
                                EZ_CURSOR,       EZ_GetCursor(XC_crosshair),
                                0);
  copy2  = EZ_CreateWidgetXrm(Locator_TYPE,    NULL,
                              "locator2",      "locator2",
                              EZ_BORDER_TYPE,  EZ_BORDER_DOWN,
                              EZ_PADB,         2,
                              EZ_BORDER_WIDTH, 4,
                              EZ_CURSOR,       EZ_GetCursor(XC_crosshair),
                              0);
  copy  = EZ_CreateWidgetXrm(Locator_TYPE,    NULL,
                             "locator3",      "locator3",
                             EZ_BORDER_TYPE,  EZ_BORDER_DOWN,
                             EZ_PADB,         2,
                             EZ_BORDER_WIDTH, 4,
                             EZ_CURSOR,       EZ_GetCursor(XC_crosshair),
                             EZ_MOTION_CALLBACK, motionCb, copy2,
                             0);
  label =    EZ_CreateWidget(EZ_WIDGET_LABEL,  frame,
                             EZ_HEIGHT,        30,
                             EZ_LABEL_STRING,  "Location",
                             EZ_PROPAGATE,     False,
                             0);

  btn =      EZ_CreateWidget(EZ_WIDGET_NORMAL_BUTTON, frame,
                             EZ_LABEL_STRING,         "Exit",
                             EZ_SIZE,                 0,0,
                             EZ_CALLBACK,             getout, NULL,
                             0);

  icon = EZ_CreateWidget(EZ_WIDGET_FREE_LABEL, NULL,
                         EZ_LABEL_STRING, "Drag Icon",
                         EZ_BACKGROUND,   "#00cccc",
                         EZ_FOREGROUND,   "red",
                         NULL);
  EZ_AddEventHandler(locator,pEHandler, label, 0);
  EZ_WidgetAddDnDDataEncoder(locator, TEST_ATOM, 0, encodeLoc, NULL, NULL, NULL);
  EZ_WidgetAddDnDDataEncoder(locator, EZ_DND_DRAG_INIT_ATOM, 0, 
                             setDragIconEncoder, icon, NULL, NULL);

  EZ_WidgetAddDnDDataDecoder(btn, TEST_ATOM, 0, decodeLoc, NULL, NULL, NULL);

  EZ_DisplayWidget(frame);
  EZ_DisplayWidget(copy);   EZ_DisplayWidget(copy2);
  EZ_EventMainLoop();
}
/*#endif*/
/*********************************************************************/

